import functools
from Bio.Align.Applications import MafftCommandline
from Bio.Align.Applications import PrankCommandline
from Bio.Align.Applications import ClustalwCommandline


aligner_params_dict = {
    "MAFFT": functools.partial(MafftCommandline, globalpair=True, maxiterate=0),
    "PRANK": functools.partial(PrankCommandline, f="fasta", once=False),
    "CLUSTALW": functools.partial(ClustalwCommandline, output="FASTA")
}


class Aligner:
    def __init__(self, aligner: str) -> None:
        self._aligner_name = aligner
        self._aligner = aligner_params_dict[aligner]

    def get_name(self):
        return self._aligner_name
    
    def set_input_file(self, file_path, tree_file=None):
        self._input_file = str(file_path)
        self._output_name = str(file_path).split(".")[0]
        if self._aligner_name == "MAFFT":
            self._aligner_cmd = self._aligner(input=self._input_file)
        if self._aligner_name == "PRANK":
            self._aligner_cmd = self._aligner(d=self._input_file, o=self._output_name, t=str(tree_file))
        if self._aligner_name == "CLUSTALW":
            self._aligner_cmd = self._aligner(infile=self._input_file, outfile=f"{self._output_name}_realigned.fasta")


    def get_realigned_msa(self) -> str:
        # print(self._aligner_cmd)
        if self._aligner_name == "MAFFT":
            realigned_msa, stderr = self._aligner_cmd()
        if self._aligner_name == "PRANK":
            realigned_msa, stderr = self._aligner_cmd()
            with open(f"{self._output_name}.best.fas",'r') as f:
                realigned_msa = f.read()
        if self._aligner_name == "CLUSTALW":
            realigned_msa, stderr = self._aligner_cmd()
            with open(f"{self._output_name}_realigned.fasta",'r') as f:
                realigned_msa = f.read()


        return realigned_msa





import pickle, pathlib, os
import numpy as np
import random

from msasim import sailfish as sf

from spartaabc.getting_priors import get_means

TRUNCATION = 150

length_distribution_priors = {
    "zipf": {
        "insertion": sorted(get_means.final_priors["zipf"]),
        "deletion": sorted(get_means.final_priors["zipf"])
    },
    "geometric": {
        "insertion": sorted(get_means.final_priors["geometric"]),
        "deletion": sorted(get_means.final_priors["geometric"])
    },
    "poisson": {
        "insertion": sorted(get_means.final_priors["poisson"]),
        "deletion": sorted(get_means.final_priors["poisson"])
    }
}


def fast_zipf(a_param, truncation=150):
    harmonic_series = np.arange(1,truncation+1)
    harmonic_series = np.power(harmonic_series, -a_param)
    harmonic_sum = np.sum(harmonic_series)
    return harmonic_series / harmonic_sum

length_dist_mapper = {
    "zipf": sf.CustomDistribution,
    "poisson": sf.PoissonDistribution,
    "geometric": sf.GeometricDistribution
}

def protocol_updater(protocol: sf.SimProtocol, params: list) -> None:
    protocol.set_sequence_size(params[0])
    protocol.set_insertion_rates(insertion_rate=params[1])
    protocol.set_deletion_rates(deletion_rate=params[2])
    protocol.set_insertion_length_distributions(insertion_dist=params[3])
    protocol.set_deletion_length_distributions(deletion_dist=params[4])

class PriorSampler:
    def __init__(self, conf_file=None,
                       len_dist="zipf",
                       rate_priors=[[-3,-1],[-1,1]], # log
                       seq_lengths=[100,500],
                       indel_model="sim",
                       seed = 1):
        self.seed = seed
        random.seed(seed)
        self.indel_model = indel_model

        self.length_distribution = length_dist_mapper[len_dist]
        self.len_dist = len_dist

        self.len_prior_dict = length_distribution_priors[len_dist]

        self.rate_prior_dict = {
            "sum_rates": rate_priors[0],
            "ratio_rates": rate_priors[1]
        }

        self.sequence_length_prior = [int(seq_lengths[0]*0.8), int(seq_lengths[1]*1.1)]

    def sample_root_length(self):
        while True:
            root_length = random.randint(*self.sequence_length_prior)
            yield root_length

    def sample_length_distributions(self):
        while True:
            x = random.uniform(*self.len_prior_dict["insertion"])
            if self.indel_model == "sim":
                indel_length_dist = self.length_distribution(fast_zipf(x, TRUNCATION))
                indel_length_dist.p = x
                yield self.len_dist, indel_length_dist, indel_length_dist
            else:
                y = random.uniform(*self.len_prior_dict["deletion"])
                indel_length_dist_insertion = self.length_distribution(fast_zipf(x, TRUNCATION))
                indel_length_dist_insertion.p = x
                indel_length_dist_deletion = self.length_distribution(fast_zipf(y, TRUNCATION))
                indel_length_dist_deletion.p = y
                yield self.len_dist, indel_length_dist_insertion, indel_length_dist_deletion



    def sample_rates(self):
        while True:
            sum_of_rates = 10**random.uniform(*self.rate_prior_dict["sum_rates"])
            ratio_of_rates = 10**random.uniform(*self.rate_prior_dict["ratio_rates"])
            if self.indel_model == "sim":
                yield (sum_of_rates, sum_of_rates)
            else:
                deletion_rate = sum_of_rates/(ratio_of_rates+1)
                insertion_rate = sum_of_rates - deletion_rate
                yield (insertion_rate, deletion_rate)


    def sample(self, n=1):
        root_length = self.sample_root_length()
        indel_rates = self.sample_rates()
        length_dists = self.sample_length_distributions()
        params_sample = []
        for params in zip(root_length, indel_rates, length_dists):
            if n==0:
                break
            params_sample.append(params)
            n = n - 1
        return params_sample


import os
import re
from pathlib import Path

from msasim import sailfish as sf

MODEL_MAPPER = {
    # Protein models
    "WAG": sf.MODEL_CODES.WAG,
    "LG": sf.MODEL_CODES.LG,
    "cpREV": sf.MODEL_CODES.CPREV45,
    "Dayhoff": sf.MODEL_CODES.DAYHOFF,
    "HIVb": sf.MODEL_CODES.HIVB,
    "HIVw": sf.MODEL_CODES.HIVW,
    "JTT": sf.MODEL_CODES.JONES,
    "mtREV": sf.MODEL_CODES.MTREV24,
    # Nucleotide models
    "JC": sf.MODEL_CODES.NUCJC,
    "HKY": sf.MODEL_CODES.HKY,
    "GTR": sf.MODEL_CODES.GTR,


}

def parse_raxmlNG_output(res_filepath):

    try:
        with open(res_filepath) as fpr:
            content = fpr.read()
        res_dict = parse_raxmlNG_content(content)
    except:
        print("Error with:", res_filepath)
        return

    return res_dict


def parse_raxmlNG_content(content):
    """
    :return: dictionary with the attributes - string typed. if parameter was not estimated, empty string
    """
    res_dict = dict.fromkeys(["ll", "pInv", "gamma", "cats",
                              "fA", "fC", "fG", "fT",
                              "subAC", "subAG", "subAT", "subCG", "subCT", "subGT",
                              "time"], "")

    # likelihood
    ll_re = re.search("Final LogLikelihood:\s+(.*)", content)
    if ll_re:
        res_dict["ll"] = ll_re.group(1).strip()
    elif re.search("BL opt converged to a worse likelihood score by", content) or re.search("failed", content):
        ll_ini = re.search("initial LogLikelihood:\s+(.*)", content)
        if ll_ini:
            res_dict["ll"] = ll_ini.group(1).strip()
    else:
        res_dict["ll"] = 'unknown raxml-ng error, check "parse_raxmlNG_content" function'


    # gamma (alpha parameter) and proportion of invariant sites
    gamma_regex = re.search("alpha:\s+(\d+\.?\d*)\s+", content)
    pinv_regex = re.search("P-inv.*:\s+(\d+\.?\d*)", content)
    cats_regex = re.search("\(\d+ cats,", content)
    if gamma_regex:
        res_dict['gamma'] = gamma_regex.group(1).strip()
    if pinv_regex:
        res_dict['pInv'] = pinv_regex.group(1).strip()
    if cats_regex:
        res_dict['cats'] = cats_regex.group(0)[1]

    # Nucleotides frequencies
    nucs_freq = re.search("Base frequencies.*?:\s+(\d+\.?\d*)\s+(\d+\.?\d*)\s+(\d+\.?\d*)\s+(\d+\.?\d*)", content)
    if nucs_freq:
        for i,nuc in enumerate("ACGT"):
            res_dict["f" + nuc] = nucs_freq.group(i+1).strip()

    # substitution frequencies
    subs_freq = re.search("Substitution rates.*:\s+(\d+\.?\d*)\s+(\d+\.?\d*)\s+(\d+\.?\d*)\s+(\d+\.?\d*)\s+(\d+\.?\d*)\s+(\d+\.?\d*)", content)
    if subs_freq:
        for i,nuc_pair in enumerate(["AC", "AG", "AT", "CG", "CT", "GT"]):  # todo: make sure order
            res_dict["sub" + nuc_pair] = subs_freq.group(i+1).strip()

    # Elapsed time of raxml-ng optimization
    rtime = re.search("Elapsed time:\s+(\d+\.?\d*)\s+seconds", content)
    if rtime:
        res_dict["time"] = rtime.group(1).strip()
    else:
        res_dict["time"] = 'no ll opt_no time'

    return res_dict




def parse_raxml_bestModel(model_path: Path):
    try:
        model_file = next(model_path.glob("*.bestModel"))
    except StopIteration as e:
        raise Exception(f"No substitution model provided in {model_path}, please specify a model using the RaxML-ng bestModel format.")

        # Initialize results dictionary
    results = {
        'submodel': None,
        'empirical_frequencies': False,
        'gamma_cats': 1,
        'gamma_shape': 1.0,
        'partition': None
    }
    
    # Split into model and partition if partition exists
    parts = model_file.read_text().split(',')
    model_part = parts[0].strip()
    
    # Parse partition information if present
    if len(parts) > 1:
        partition_part = parts[1].strip()
        results['partition'] = partition_part
    
    # Split model into components
    model_components = model_part.split('+')
    
    # First component is always the substitution model
    results['submodel'] = MODEL_MAPPER.get(model_components[0], -1)
    if (results['submodel']) == -1:
        raise RuntimeError("The requested model has not been implemented :(")

    
    # Parse remaining components
    for component in model_components[1:]:
        # Check for empirical frequencies
        if component == 'FC':
            results['empirical_frequencies'] = True
        
        # Check for gamma categories and alpha
        elif component.startswith('G'):
            # Extract number of categories
            gamma_info = component.split('m')
            results['gamma_cats'] = int(gamma_info[0][1:])  # Remove 'G' and convert to int
            
            # Extract alpha if present
            if len(gamma_info) > 1:
                alpha_str = gamma_info[1].strip('{}')
                results['gamma_shape'] = float(alpha_str)
    
    return results
    



def get_substitution_model(path):
    return parse_raxml_bestModel(path)
    # joined_path = os.path.join(path,"T1.raxml.log")

    # res_dict = parse_raxmlNG_output(joined_path)

    # x,y,z,m,n,k = res_dict["subAC"],res_dict["subAG"],res_dict["subAT"],res_dict["subCG"],res_dict["subCT"], res_dict["subGT"]
    # x,y,z,m,n,k = float(x),float(y),float(z),float(m),float(n),float(k)
    # pi_A,pi_C,pi_G,pi_T = res_dict["fA"],res_dict["fC"],res_dict["fG"],res_dict["fT"]
    # pi_A,pi_C,pi_G,pi_T = float(pi_A),float(pi_C),float(pi_G),float(pi_T)

    # a,b,c,d,e = (n/y)*(pi_G/pi_T), (z/y)*(pi_G/pi_T), (1.0/y)*(pi_G/pi_T), (x/y)*(pi_G/pi_C), (m/y)

    # subtitution_model = {
    #     "freq": (pi_A, pi_C, pi_G, pi_T),
    #     "rates": (x,y,z,m,n,k),
    #     "inv_prop": res_dict["pInv"],
    #     "gamma_shape": res_dict["gamma"],
    #     "gamma_cats": res_dict["cats"],
    #     "submodel": "GTR"
    # }
    # return subtitution_model


import numpy as np
import pandas as pd


from scipy.optimize import fsolve


def calc_zip_mom(a, mom=1, truncation=50, cutoff=None):
    """
    Calculation of Zipfian moment
    inputs:
    a - a parameter of the Zipfian distribution
    mom - (optional, by default 1) the moment you would like to calculate (for example mom=1 -> average)
    truncation - (optional, by default 150) maximal value of the Zipfian distribution
    cutoff - (optional, by default None) if not none it will give the sum for the distribution until cutoff
    output:
    moment of the distribution (if cutoff!=None it will return the partial sum
    """
    if a <= 1:
        return 10000000
    a = float(a)
    mom = int(mom)
    truncation = int(truncation)
    z = 0
    out = 0
    for i in range(1, truncation + 1, 1):
        z += i ** -a
        if (cutoff is None) or (cutoff is not None and i <= cutoff):
            out += (i ** -a) * (i ** mom)
    return out / z


def zip_mom_to_a(mom, truncation=50, init_guess=1.00000001, epsfcn=0.0000001, xtol=1.49012e-10):
    """
    Returns the "a" parameter of the truncated zipfian function with the input moment
    mom - value of the moment
    mom_order - order of the moment (1 is the average)
    truncation -  where the distribution is truncated
    """
    return fsolve(lambda x: (mom - calc_zip_mom(x, mom=1, truncation=truncation)) \
                  , init_guess, epsfcn=epsfcn, xtol=xtol)[0]


import logging, argparse, subprocess, sys
from pathlib import Path

from spartaabc.utility import logger, setLogHandler

interpreter=sys.executable

def parse_args(arg_list: list[str] | None):
    _parser = argparse.ArgumentParser(allow_abbrev=False)
    _parser.add_argument('-i','--input', action='store',metavar="Input folder", type=str, required=True)
    # _parser.add_argument('-c','--config', action='store',metavar="Simulation config" , type=str, required=True)
    _parser.add_argument('-t','--type', action='store',metavar="Type of MSA NT/AA" , type=str, required=True)
    _parser.add_argument('-n','--numsim', action='store',metavar="Number of simulations" , type=int, required=True)
    _parser.add_argument('-nc','--numsim-correction', action='store',metavar="Number of correction simulations" , type=int, required=True)

    _parser.add_argument('-s','--seed', action='store',metavar="Simulator seed" , type=int, required=True)
    _parser.add_argument('-a','--aligner', action='store',metavar="Alignment program to use" , type=str, default="mafft", required=False)

    _parser.add_argument('-k','--keep-stats', action='store_true')
    _parser.add_argument('-v','--verbose', action='store_true')


    args = _parser.parse_args()
    return args

def run_pipeline_parallel(MAIN_PATH: Path, SEED: int, SEQUENCE_TYPE: str,
                          NUM_SIMS: int, NUM_SIMS_CORRECTION: int, INDEL_MODELS: list[str],
                          ALIGNER: str, KEEP_STATS: bool):
    logging.basicConfig()

    CURRENT_SCRIPT_DIR = Path(__file__).parent
    print(CURRENT_SCRIPT_DIR)


    print()

    setLogHandler(MAIN_PATH, "w")
    logger.info("\n\tMAIN_PATH: {},\n\tSEED: {}, NUM_SIMS: {}, NUM_SIMS_CORRECTION: {}, SEQUENCE_TYPE: {}".format(
        MAIN_PATH, SEED, NUM_SIMS, NUM_SIMS_CORRECTION, SEQUENCE_TYPE
    ))


    processes = []
    for model in INDEL_MODELS:
        simulate_cmd = [interpreter, CURRENT_SCRIPT_DIR / "simulate_data.py",
                        "-i", str(MAIN_PATH), "-n", str(NUM_SIMS),
                        "-s", str(SEED), "-l", "zipf", "-m", f"{model}"]
    
        correction_cmd_sim = [interpreter, CURRENT_SCRIPT_DIR / "correction.py",
                              "-i", str(MAIN_PATH), "-n", str(NUM_SIMS_CORRECTION),
                              "-s", str(SEED+1), "-l", "zipf", "-m", f"{model}",
                              "-t", SEQUENCE_TYPE]
        SEED += 2

        processes.append(subprocess.Popen(simulate_cmd))
        processes.append(subprocess.Popen(correction_cmd_sim))

    exit_codes = [p.wait() for p in processes]
    
    abc_cmd = [interpreter, CURRENT_SCRIPT_DIR / "abc_inference.py", "-i", str(MAIN_PATH)]
    subprocess.run(abc_cmd)



def main(arg_list: list[str] | None = None):
    logging.basicConfig()

    CURRENT_SCRIPT_DIR = Path(__file__).parent
    print(CURRENT_SCRIPT_DIR)
    args = parse_args(arg_list)

    MAIN_PATH = Path(args.input).resolve()
    SEED = args.seed
    SEQUENCE_TYPE = args.type
    NUM_SIMS = args.numsim
    NUM_SIMS_CORRECTION = args.numsim_correction

    ALIGNER = args.aligner.upper()
    KEEP_STATS = args.keep_stats
    VERBOSE = args.verbose

    print()

    setLogHandler(MAIN_PATH, "w")
    logger.info("\n\tMAIN_PATH: {},\n\tSEED: {}, NUM_SIMS: {}, NUM_SIMS_CORRECTION: {}, SEQUENCE_TYPE: {}".format(
        MAIN_PATH, SEED, NUM_SIMS, NUM_SIMS_CORRECTION, SEQUENCE_TYPE
    ))

    INDEL_MODELS = ["sim", "rim"]

    processes = []
    for model in INDEL_MODELS:
        simulate_cmd = [interpreter, CURRENT_SCRIPT_DIR / "simulate_data.py",
                        "-i", str(MAIN_PATH), "-n", str(NUM_SIMS),
                        "-s", str(SEED), "-l", "zipf", "-m", f"{model}"]
    
        correction_cmd_sim = [interpreter, CURRENT_SCRIPT_DIR / "correction.py",
                              "-i", str(MAIN_PATH), "-n", str(NUM_SIMS_CORRECTION),
                              "-s", str(SEED+1), "-l", "zipf", "-m", f"{model}",
                              "-t", SEQUENCE_TYPE]
        SEED += 2

        processes.append(subprocess.Popen(simulate_cmd))
        processes.append(subprocess.Popen(correction_cmd_sim))

    exit_codes = [p.wait() for p in processes]
    
    abc_cmd = [interpreter, CURRENT_SCRIPT_DIR / "abc_inference.py", "-i", str(MAIN_PATH)]
    subprocess.run(abc_cmd)




if __name__=="__main__":
    main()


from spartaabc.getting_priors import geometric as geo
from spartaabc.getting_priors import poisson as pois
from spartaabc.getting_priors import zipf


length_distributions = ["zipf", "geometric", "poisson"]


params_search_map = {
    "geometric" :  geo.geo_moment_to_p,
    "zipf":  zipf.zip_mom_to_a,
    "poisson": pois.pois_moment_to_mu
}

mean_check_map = {
    "geometric" :  geo.get_moment_geoemtric,
    "zipf":  zipf.calc_zip_mom,
    "poisson": pois.get_moment_poisson
}

TRUNCATION = 150

lower_mean = 1.5
upper_mean = 25.0

lower_prior_lims = {}
upper_prior_lims = {}


epsilon = 0.001

for len_dist in length_distributions:
    # calculate prior limits (params) for given means:
    lower_prior_lims[len_dist] = params_search_map[len_dist](lower_mean, TRUNCATION)    
    upper_prior_lims[len_dist] = params_search_map[len_dist](upper_mean, TRUNCATION)    

    # get the resulting means from calculated params:
    lower_mean_test = mean_check_map[len_dist](lower_prior_lims[len_dist], 1, TRUNCATION)
    upper_mean_test = mean_check_map[len_dist](upper_prior_lims[len_dist], 1, TRUNCATION)

    # test if calculated param is correct within epsilon threshold:
    if (lower_mean_test < lower_mean - epsilon) or (lower_mean > lower_mean + epsilon):
        print(f"Could not find lower mean param for {len_dist}")

    if (upper_mean_test < upper_mean - epsilon) or (upper_mean_test > upper_mean + epsilon):
        print(f"Could not find upper mean param for {len_dist}")

# organize the results for printing:
priors = tuple(zip(lower_prior_lims.values(), upper_prior_lims.values()))
final_priors = dict(zip(length_distributions, priors))

# print(final_priors)


import os
import tempfile
import pickle
import argparse
import warnings
import copy
import logging
from pathlib import Path
import numpy as np
import pandas as pd
from sklearn import linear_model, model_selection, exceptions
from sklearn.pipeline import Pipeline
from scipy.stats import pearsonr

from msasim import sailfish as sf
import msastats

from spartaabc.prior_sampler import PriorSampler, protocol_updater
from spartaabc.aligner_interface import Aligner
from spartaabc.raxml_parser import get_substitution_model
from spartaabc.utility import get_msa_path, get_tree_path, prepare_prior_sampler, setLogHandler
from spartaabc.utility import StandardMemoryScaler, logger

warnings.simplefilter("ignore", category=exceptions.ConvergenceWarning)

def parse_args(arg_list: list[str] | None):
    _parser = argparse.ArgumentParser(allow_abbrev=False)
    _parser.add_argument('-i','--input', action='store',metavar="Input folder", type=str, required=True)
    # _parser.add_argument('-c','--config', action='store',metavar="Simulation config" , type=str, required=True)
    _parser.add_argument('-t','--type', action='store',metavar="Type of MSA NT/AA" , type=str, required=True)
    _parser.add_argument('-n','--numsim', action='store',metavar="Number of simulations" , type=int, required=True)
    _parser.add_argument('-s','--seed', action='store',metavar="Simulator seed" , type=int, required=True)
    _parser.add_argument('-a','--aligner', action='store',metavar="Alignment program to use" , type=str, default="mafft", required=False)

    _parser.add_argument('-l','--lengthdist', action='store',metavar="Simulation config" , type=str, required=True)
    _parser.add_argument('-m','--model', action='store',metavar="Simulation config" , type=str, required=True)
    _parser.add_argument('-k','--keep-stats', action='store_true')
    _parser.add_argument('-v','--verbose', action='store_true')


    args = _parser.parse_args()
    return args



# prepare indelible control file for subsitutions:
def prepare_substitution_model(main_path: Path, sequence_type: str):

    substitution_model = get_substitution_model(main_path)# if sequence_type == "NT" else {}
    substitution_model["mode"] = "DNA" if sequence_type == "NT" else "PROTEIN"
    
    return substitution_model


def simulate_data(prior_sampler: PriorSampler, num_sims: int, tree_path: str, substitution_model: dict, seed: int):
    print(seed)
    sim_protocol = sf.SimProtocol(tree_path, seed=seed)
    simulator = sf.Simulator(sim_protocol,
                             simulation_type=sf.SIMULATION_TYPE[substitution_model["mode"]])

    simulated_msas = []
    sum_stats = []

    sim_params_correction = prior_sampler.sample(num_sims)

    simulator.set_replacement_model(model=substitution_model["submodel"],
                                    model_parameters=substitution_model.get("params", None),
                                    gamma_parameters_alpha=substitution_model.get("gamma_shape", 1.0),
                                    gamma_parameters_categories=substitution_model.get("gamma_cats", 1))

    logger.info("Starting msa simulation")
    for idx,params in enumerate(sim_params_correction):
        root_length = params[0]
        insertion_rate, deletion_rate = params[1]
        lendist, insertion_length_dist, deletion_length_dist = params[2]

        numeric_params = [root_length, insertion_rate, deletion_rate, insertion_length_dist.p, deletion_length_dist.p]
        protocol_updater(sim_protocol, [root_length, insertion_rate, deletion_rate,
                         insertion_length_dist, deletion_length_dist])

        sim_msa = simulator()
        sim_stats = msastats.calculate_msa_stats(sim_msa.get_msa().splitlines()[1::2])
        # print(sim_stats)
        simulated_msas.append(sim_msa)
        sum_stats.append(numeric_params + sim_stats)
    logger.info(f"Done with {num_sims} msa simulations")

    return simulated_msas, sum_stats


def compute_realigned_stats(msas: list[sf.Msa], sum_stats: list[list[float]],
                            sequence_aligner: Aligner, tree_path: str):
    logger.info("Realigning MSAs and recomputing stats")
    realigned_sum_stats = []

    for msa in msas:
        sim_fasta_unaligned = msa.get_msa().replace("-","").encode()
        with tempfile.NamedTemporaryFile(suffix='.fasta') as tempf:
            tempf.write(sim_fasta_unaligned)
            tempf.seek(0)
            sequence_aligner.set_input_file(tempf.name, tree_file=tree_path)
            realigned_msa = sequence_aligner.get_realigned_msa()
            
        realigned_msa = [s[s.index("\n"):].replace("\n","") for s in realigned_msa.split(">")[1:]]
        realigned_stats = msastats.calculate_msa_stats(realigned_msa)
        realigned_sum_stats.append(realigned_stats)
    
    logger.info("Done recomputing stats")
    return realigned_sum_stats

def compute_regressors(true_stats: list[list[float]], corrected_stats: list[list[float]]):
    logger.info("Performing regression on all realigned stats")

    X = np.array(true_stats, dtype=float)
    Y = np.array(corrected_stats, dtype=float).T

    reg = linear_model.Lasso()
    parameters = {'alpha':np.logspace(-7,4,20)}
    clf_lassocv = model_selection.GridSearchCV(estimator = reg,
                                param_grid = parameters, cv=3,
                                scoring = 'neg_mean_squared_error')
    regression_pipline = Pipeline([("scaler", StandardMemoryScaler()),('regression', clf_lassocv)])
    regressors = []
    performance_metrics = []
    for y in Y:
        regression_pipline.fit(X, y)
        saved_estimator = copy.deepcopy(regression_pipline)
        regressors.append(saved_estimator)
        
        Y_pred = regression_pipline.predict(X)
        r_val, p_val = pearsonr(Y_pred,y)
        performance_metrics.append({
            'pearsonr': r_val,
            'p_val': p_val,
            'mean_test_score': np.min(np.sqrt(-clf_lassocv.cv_results_['mean_test_score']))
        })
    logger.info("Done with regression")
    return regressors, performance_metrics


def compute_alignment_correction(MAIN_PATH: Path, SEED: int, SEQUENCE_TYPE: str,
                                 NUM_SIMS: int, ALIGNER: str, LENGTH_DISTRIBUTION: str,
                                 INDEL_MODEL: str, KEEP_STATS: bool):
    logging.basicConfig()

    ALIGNER = Aligner(ALIGNER.upper())

    setLogHandler(MAIN_PATH)
    logger.info("\n\tMAIN_PATH: {},\n\tSEED: {}, NUM_SIMS: {}, SEQUENCE_TYPE: {},\n\tLENGTH_DISTRIBUTION: {}, INDEL_MODEL {}".format(
        MAIN_PATH, SEED, NUM_SIMS, SEQUENCE_TYPE, LENGTH_DISTRIBUTION, INDEL_MODEL
    ))


    TREE_PATH = get_tree_path(MAIN_PATH)
    MSA_PATH = get_msa_path(MAIN_PATH)

    prior_sampler = prepare_prior_sampler(MSA_PATH, LENGTH_DISTRIBUTION, INDEL_MODEL, SEED)
    substitution_model = prepare_substitution_model(MAIN_PATH, SEQUENCE_TYPE)
    msas, sum_stats = simulate_data(prior_sampler=prior_sampler, num_sims=NUM_SIMS,
                                    tree_path=TREE_PATH, substitution_model=substitution_model,
                                    seed=SEED)
    realigned_sum_stats = compute_realigned_stats(msas, sum_stats, ALIGNER, TREE_PATH)
    regressors, regressors_performence = compute_regressors(true_stats=sum_stats, corrected_stats=realigned_sum_stats)

    full_correction_path = MAIN_PATH / f"{ALIGNER.get_name()}_correction"
    try:
        os.mkdir(full_correction_path)
    except:
        print("correction folder exists already")
    
    with open(full_correction_path / f'regressors_{LENGTH_DISTRIBUTION}_{INDEL_MODEL}.pickle', 'wb') as f:
        pickle.dump(regressors, f)
    pd.DataFrame(regressors_performence).to_csv(
        full_correction_path / f'regression_performance_{LENGTH_DISTRIBUTION}_{INDEL_MODEL}.csv')

    if KEEP_STATS:
        print("saving stats...")
        true_stats = pd.DataFrame(sum_stats)
        true_stats.columns = map(str, range(len(true_stats.columns)))
        true_stats.to_parquet(full_correction_path / "true_stats.parquet.gzip", compression='gzip', index=False)

        realigned_stats = pd.DataFrame(realigned_sum_stats)
        realigned_stats.columns = map(str, realigned_stats.columns)
        realigned_stats.to_parquet(full_correction_path / "realigned_stats.parquet.gzip", compression='gzip', index=False)

        infered_stats = np.array([regressor.predict(true_stats.values).T for regressor in regressors])
        infered_stats = pd.DataFrame(infered_stats.T, columns=map(str, range(27)))
        infered_stats.to_parquet(full_correction_path / "infered_stats.parquet.gzip", compression='gzip', index=False)


def main(arg_list: list[str] | None = None):
    logging.basicConfig()
    args = parse_args(arg_list)

    MAIN_PATH = Path(args.input).resolve()
    SEED = args.seed
    SEQUENCE_TYPE = args.type
    NUM_SIMS = args.numsim
    ALIGNER = Aligner(args.aligner.upper())
    LENGTH_DISTRIBUTION = args.lengthdist
    INDEL_MODEL = args.model
    KEEP_STATS = args.keep_stats
    VERBOSE = args.verbose

    setLogHandler(MAIN_PATH)
    logger.info("\n\tMAIN_PATH: {},\n\tSEED: {}, NUM_SIMS: {}, SEQUENCE_TYPE: {},\n\tLENGTH_DISTRIBUTION: {}, INDEL_MODEL {}".format(
        MAIN_PATH, SEED, NUM_SIMS, SEQUENCE_TYPE, LENGTH_DISTRIBUTION, INDEL_MODEL
    ))


    TREE_PATH = get_tree_path(MAIN_PATH)
    MSA_PATH = get_msa_path(MAIN_PATH)

    prior_sampler = prepare_prior_sampler(MSA_PATH, LENGTH_DISTRIBUTION, INDEL_MODEL, SEED)
    substitution_model = prepare_substitution_model(MAIN_PATH, SEQUENCE_TYPE)
    msas, sum_stats = simulate_data(prior_sampler=prior_sampler, num_sims=NUM_SIMS,
                                    tree_path=TREE_PATH, substitution_model=substitution_model,
                                    seed=SEED)
    realigned_sum_stats = compute_realigned_stats(msas, sum_stats, ALIGNER, TREE_PATH)
    regressors, regressors_performence = compute_regressors(true_stats=sum_stats, corrected_stats=realigned_sum_stats)

    full_correction_path = MAIN_PATH / f"{args.aligner}_correction"
    try:
        os.mkdir(full_correction_path)
    except:
        print("correction folder exists already")
    
    with open(full_correction_path / f'regressors_{LENGTH_DISTRIBUTION}_{INDEL_MODEL}.pickle', 'wb') as f:
        pickle.dump(regressors, f)
    pd.DataFrame(regressors_performence).to_csv(
        full_correction_path / f'regression_performance_{LENGTH_DISTRIBUTION}_{INDEL_MODEL}.csv')

    if KEEP_STATS:
        print("saving stats...")
        true_stats = pd.DataFrame(sum_stats)
        true_stats.columns = map(str, range(len(true_stats.columns)))
        true_stats.to_parquet(full_correction_path / "true_stats.parquet.gzip", compression='gzip', index=False)

        realigned_stats = pd.DataFrame(realigned_sum_stats)
        realigned_stats.columns = map(str, realigned_stats.columns)
        realigned_stats.to_parquet(full_correction_path / "realigned_stats.parquet.gzip", compression='gzip', index=False)

        infered_stats = np.array([regressor.predict(true_stats.values).T for regressor in regressors])
        infered_stats = pd.DataFrame(infered_stats.T, columns=map(str, range(27)))
        infered_stats.to_parquet(full_correction_path / "infered_stats.parquet.gzip", compression='gzip', index=False)



if __name__ == '__main__':
    main()


import argparse
import pickle
import logging
from pathlib import Path
import numpy as np
import pandas as pd
from dataclasses import dataclass

import msastats
from spartaabc.utility import get_msa_path
from spartaabc.utility import PARAMS_LIST, SUMSTATS_LIST
from spartaabc.utility import logger, setLogHandler



@dataclass
class IndelParams:
    root_length: int
    insertion_rate: float
    deletion_rate: float
    insertion_length_parameter: float
    deletion_length_parameter: float
    length_distribution: str
    indel_model: str

    def __repr__(self):
        model_str = f"Model: {self.indel_model}\n"
        model_str += f"Root_length: {self.root_length}\n"
        if self.indel_model in ["SIM", "sim"]:
            model_str += f"R_ID: {self.insertion_rate}\n"
            model_str += f"A_ID: {self.insertion_length_parameter}"
        elif self.indel_model in ["RIM", "rim"]:
            model_str += f"R_I: {self.insertion_rate}\n"
            model_str += f"R_D: {self.deletion_rate}\n"
            model_str += f"A_I: {self.insertion_length_parameter}\n"
            model_str += f"A_D: {self.deletion_length_parameter}"

        return model_str

def parse_args(arg_list: list[str] | None):
    _parser = argparse.ArgumentParser(allow_abbrev=False)
    _parser.add_argument('-i','--input', action='store',metavar="Input folder", type=str, required=True)
    _parser.add_argument('-a','--aligner', action='store',metavar="Aligner", type=str,default="mafft" , required=False)
    _parser.add_argument('-d','--distance', action='store',metavar="Distance metric", type=str, default="mahal", required=False)


    args = _parser.parse_args()
    return args


def load_data(main_path: Path):
    full_data = {}
    for data_path in main_path.glob("*.parquet.gzip"):
        model = data_path.stem.split('.')[0].split("_", maxsplit=2)[2]
        temp_df = pd.read_parquet(data_path)
        full_data[model] = temp_df

    return full_data

def load_correction_regressors(main_path: Path, aligner: str):
    regressors = {}
    for regressor_path in (main_path / f"{aligner}_correction").glob("*.pickle"):
        model = regressor_path.stem.split("_", maxsplit=1)[1]
        with open(regressor_path, 'rb') as f:
            regressors[model] = pickle.load(f)
    return regressors

def load_correction_regressor_scores(main_path: Path, aligner: str):
    scores = pd.DataFrame(len(SUMSTATS_LIST)*[1.0], columns=["pearsonr"])
    for score_path in (main_path / f"{aligner}_correction").glob("*.csv"):
        score_df = pd.read_csv(score_path, index_col=0)[["pearsonr"]]
        scores[scores["pearsonr"] > score_df["pearsonr"]] = score_df

    return scores["pearsonr"].to_list()

def bias_correction(regressors, data: pd.DataFrame, regressor_scores: list[float]):
    data = data.to_numpy()

    kept_stats = []
    infered_data = []
    for idx, regressor in enumerate(regressors):
        if regressor_scores[idx] > 0.80:
            kept_stats.append(idx)
            infered_data.append(regressor.predict(data).T)

    temp_data = np.array(infered_data)
    temp_data = pd.DataFrame(temp_data.T, columns=[SUMSTATS_LIST[i] for i in kept_stats])
    return temp_data, kept_stats

def run(main_path: Path, aligner: str, distance_metric: str="mahal", top_cutoff: int=100) -> IndelParams:


    MSA_PATH = get_msa_path(main_path)

    empirical_stats = msastats.calculate_fasta_stats(MSA_PATH)

    stats_data = load_data(main_path)
    regressors = load_correction_regressors(main_path, aligner)
    regressor_scores = load_correction_regressor_scores(main_path, aligner)

    params_data = []
    full_stats_data = []
    for model in  stats_data.keys():
        current_regressors = regressors.get(model, None)
        params_data.append(stats_data[model][PARAMS_LIST])

        if current_regressors is not None:
            temp_df, kept_statistics = bias_correction(current_regressors, stats_data[model], regressor_scores)
            full_stats_data.append(temp_df)
    
    empirical_stats = [empirical_stats[i] for i in kept_statistics]

    params_data = pd.concat(params_data)
    full_stats_data = pd.concat(full_stats_data)

    calculated_distances = None

    if distance_metric == "mahal":
        cov = np.cov(full_stats_data.T)
        cov = cov + np.eye(len(cov))*1e-4
        inv_covmat = np.linalg.inv(cov)
        u_minus_v = empirical_stats-full_stats_data
        left = np.dot(u_minus_v, inv_covmat)
        calculated_distances = np.sqrt(np.sum(u_minus_v*left, axis=1))
    if distance_metric == "euclid":
        weights = 1/(full_stats_data.std(axis=0) + 0.001)
        calculated_distances = np.sum(weights*(full_stats_data - empirical_stats)**2, axis=1)
    
    full_stats_data["distances"] = calculated_distances
    full_stats_data[PARAMS_LIST] = params_data

    top_stats = full_stats_data.nsmallest(top_cutoff, "distances")

    top_stats[["distances"] + PARAMS_LIST].to_csv(main_path / "top_params.csv", index=False)

    abc_indel_params = None
    if len(top_stats[top_stats["insertion_rate"] == top_stats["deletion_rate"]]) > (top_cutoff // 2):
        full_sim_data = full_stats_data[full_stats_data["insertion_rate"] == full_stats_data["deletion_rate"]]
        top_sim_data = full_sim_data.nsmallest(top_cutoff, "distances")
        root_length = int(top_sim_data["root_length"].mean())
        R_ID = float(top_sim_data["insertion_rate"].mean())
        A_ID = float(top_sim_data["length_param_insertion"].mean())
        abc_indel_params = IndelParams(root_length,
                                       R_ID, R_ID,
                                       A_ID, A_ID,
                                       length_distribution="zipf",
                                       indel_model="SIM")
    else:
        full_rim_data = full_stats_data[full_stats_data["insertion_rate"] != full_stats_data["deletion_rate"]]
        top_rim_data = full_rim_data.nsmallest(top_cutoff, "distances")
        root_length = int(top_rim_data["root_length"].mean())
        R_I = float(top_rim_data["insertion_rate"].mean())
        R_D = float(top_rim_data["deletion_rate"].mean())
        A_I = float(top_rim_data["length_param_insertion"].mean())
        A_D = float(top_rim_data["length_param_deletion"].mean())
        abc_indel_params = IndelParams(root_length,
                                       R_I, R_D,
                                       A_I, A_D,
                                       length_distribution="zipf",
                                       indel_model="RIM")
    (main_path / "model_params.txt").write_text(str(abc_indel_params))


def main(arg_list: list[str] | None = None):
    logging.basicConfig()

    args = parse_args(arg_list)


    MAIN_PATH = Path(args.input).resolve()
    ALIGNER = args.aligner
    DISTANCE_METRIC = args.distance

    setLogHandler(MAIN_PATH)
    logger.info("\n\tMAIN_PATH: {}".format(
        MAIN_PATH
    ))

    run(MAIN_PATH, ALIGNER, DISTANCE_METRIC)


if __name__ == "__main__":
    main()


import numpy as np
from scipy.optimize import least_squares
from scipy.stats import poisson


def get_moment_poisson(mu, mom, truncation):
    """
    Calculation of geoemtric moment
    inputs:
    p - p parameter of the geoemtric distribution
    mom - (optional, by default 1) the moment you would like to calculate (for example mom=1 -> average)
    n_trunc - (optional, by default 150) maximal value of the geoemtric distribution
    cutoff - (optional, by default None) if not none it will give the sum for the distribution until cutoff
    output:
    moment of the distribution (if cutoff!=None it will return the partial sum
    """
    CDF = lambda x: poisson.cdf(x, mu)
    
    norm_factor = CDF(truncation)



    moment = 0
    for i in range(1, truncation+1):
        prob_i = poisson.pmf(i, mu)
        moment += (prob_i/norm_factor)*(i**mom)

    return moment


def pois_moment_to_mu(mom, truncation):
    solution = least_squares(lambda x: (mom - get_moment_poisson(x, 1, truncation)),
                        x0=mom, bounds=(1, 100)).x[0]

    return solution


# m = get_moment_poisson(1.5, 1, 150)
# print(m)



import random
from pathlib import Path
from dataclasses import dataclass

from spartaabc import simulate_data
from spartaabc import correction
from spartaabc import abc_inference
from spartaabc.abc_inference import IndelParams
from spartaabc.main import run_pipeline_parallel

@dataclass
class IndelModel:
    model: str
    length_distribution: str

def prepare_data_for_inference(data_path: Path, sequnce_type: str="AA", 
                               number_of_simulations: int=100000, number_of_correction_simulations: int=500,
                               seed: int=42, aligner: str="mafft", keep_correction_data: bool=False,
                               indel_model: IndelModel=IndelModel("sim", "zipf"),
                               verbose: int=0) -> None:

    # simulate_data.generate_summary_statistics(data_path, seed, number_of_simulations, 
    #                                           indel_model.length_distribution, indel_model.model)
    
    correction.compute_alignment_correction(data_path, seed, sequnce_type, number_of_correction_simulations,
                                            aligner, indel_model.length_distribution,
                                            indel_model.model, keep_correction_data)
    

def infer_indel_params(data_path: Path, aligner: str="MAFFT", 
                      distance_metric: str="mahal", top_cutoff: int=100) -> IndelParams:
    indel_params =  abc_inference.run(data_path, aligner, distance_metric, top_cutoff)

    print(indel_params)



def standard_inference(data_path: Path, seed=random.randint(1, 1e6)):
    prepare_data_for_inference(data_path=data_path, number_of_simulations=100_000,
                            number_of_correction_simulations=1000, seed=seed, keep_correction_data=True)
    
    prepare_data_for_inference(data_path=data_path, 
                            number_of_simulations=100_000, 
                            number_of_correction_simulations=1000, 
                            indel_model=IndelModel(model="rim", length_distribution="zipf"),
                            seed=seed+2, keep_correction_data=True)

    infer_indel_params(data_path=data_path)

def parallelized_inference(data_path: Path, SEQUENCE_TYPE: str,
                           NUM_SIMS: int, NUM_SIMS_CORRECTION: int,
                           INDEL_MODELS: list[str], ALIGNER:str,
                           KEEP_STATS: bool, seed=random.randint(1, 1e6)):
    
    run_pipeline_parallel(data_path, SEED=seed, SEQUENCE_TYPE=SEQUENCE_TYPE,
                          NUM_SIMS=NUM_SIMS, NUM_SIMS_CORRECTION=NUM_SIMS_CORRECTION,
                          INDEL_MODELS=INDEL_MODELS, ALIGNER=ALIGNER,
                          KEEP_STATS=KEEP_STATS)
